@@ -1,11 +1,10 @@
-/*
- *****************************************************************************
+/**
+ * \file lib/fko_encryption.c
  *
- * File:    fko_encryption.c
- *
- * Purpose: Set/Get the spa encryption type.
- *
- *  Fwknop is developed primarily by the people listed in the file 'AUTHORS'.
+ * \brief Set/Get the spa encryption type.
+ */
+
+/*  Fwknop is developed primarily by the people listed in the file 'AUTHORS'.
  *  Copyright (C) 2009-2015 fwknop developers and contributors. For a full
  *  list of contributors, see the file 'CREDITS'.
  *
@@ -38,6 +37,20 @@
   #include "gpgme_funcs.h"
   #if HAVE_SYS_STAT_H
     #include <sys/stat.h>
+    /* Taken from cURL for win32 build
+    */
+    #if !defined(S_ISREG) && defined(S_IFMT) && defined(S_IFREG)
+        #define S_ISREG(m) (((m) & S_IFMT) == S_IFREG)
+    #endif
+    #if !defined(S_ISDIR) && defined(S_IFMT) && defined(S_IFDIR)
+        #define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
+    #endif
+    #if !defined(S_IFLNK)
+        #define S_IFLNK 0xA000
+    #endif
+    #if !defined(S_ISLNK) && defined(S_IFMT)
+        #define S_ISLNK(m) (((m) & S_IFMT) == S_IFLNK)
+    #endif
   #endif
 #endif
 
@@ -315,11 +328,11 @@ gpg_encrypt(fko_ctx_t ctx, const char *enc_key)
     /* Make a bucket big enough to hold the enc msg + digest (plaintext)
      * and populate it appropriately.
     */
-    plain = calloc(1, ctx->encoded_msg_len + ctx->digest_len + 2);
+    plain = calloc(1, pt_len);
     if(plain == NULL)
         return(FKO_ERROR_MEMORY_ALLOCATION);
 
-    pt_len = snprintf(plain, pt_len+1, "%s:%s", ctx->encoded_msg, ctx->digest);
+    pt_len = snprintf(plain, pt_len, "%s:%s", ctx->encoded_msg, ctx->digest);
 
     if(! is_valid_pt_msg_len(pt_len))
     {
@@ -463,11 +476,11 @@ gpg_decrypt(fko_ctx_t ctx, const char *dec_key)
         if(res != FKO_SUCCESS) /* bail if there was some other problem */
             return(res);
 
-    pt_len = strnlen(ctx->encoded_msg, MAX_SPA_ENCODED_MSG_SIZE);
-
     if(ctx->encoded_msg == NULL)
         return(FKO_ERROR_INVALID_DATA_ENCRYPT_DECRYPTED_MESSAGE_MISSING);
 
+    pt_len = strnlen(ctx->encoded_msg, MAX_SPA_ENCODED_MSG_SIZE);
+
     if(! is_valid_encoded_msg_len(pt_len))
         return(FKO_ERROR_INVALID_DATA_ENCRYPT_DECRYPTED_MSGLEN_VALIDFAIL);
 
